//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Maze generation: Kruskal algorithm
//
// -------------------------------------------------------------------
//
// $Id: buildK.cxx,v 1.2 2005/01/21 07:50:24 loreti Exp $

#include <algorithm>
#include <functional>
#include <iostream>
#include <list>
#include "maze.hh"

using std::vector;

// Algorithm: a room may be visited or unvisited, and at the beginning
// all of them are set unvisited.  A (non border) wall is chosen
// randomly, and the two neighbor rooms are considered.  If they are
// both unvisited, they are connected together by a (labeled) path,
// and both rooms marked visited.  If one room is visited and the
// other unvisited, they are connected together on the same path of
// the visited one.  If both are visited and on the same path, nothing
// is done; if both are visited but they are on different paths, they
// are connected together and both paths merged together.  The process
// is iterated over all the walls.

// A function object used to store a sequence of numbers

class stepper {
private:
  int first;
  int delta;

public:
  stepper(int f, int d) : first(f), delta(d) {}
  int operator () () { int i = first;  first += delta;  return i; }

};  // class stepper

void maze::buildK() {
  std::cout << "Kruskal algorithm requested to generate a "
            << _width << "-by-" << _height << " maze" << std::endl;

  // A vector of integers, filled with the numbers from 0 to (number
  // of walls minus one) and randomly shuffled: used to access all the
  // walls in a random sequence

  vector<int> wallIndex(_wallDim);
  std::generate_n(wallIndex.begin(), _wallDim, stepper(0, 1));
  std::random_shuffle(wallIndex.begin(), wallIndex.end());

  // A sequence of rooms (a "path"), a vector holding the paths
  // currently built, and the index, in the vector, of the next path

  typedef std::list<room *> path;
  vector<path> vPath;
  vector<path>::size_type nextEntry = 0;
  vector<int>::const_iterator it, finis = wallIndex.end();

  for (it = wallIndex.begin();   it != finis;   ++it
  ) {
    // Gets a wall, skipping borders

    wall *pW = _wall + *it;
    if (pW->getType() == wall::BORDER) continue;

    // Gets the two neighbor rooms and their path numbers (their
    // "flag")

    room *pBef = pW->getBefore();
    room *pAft = pW->getAfter();
    int here   = pBef->getFlag();
    int there  = pAft->getFlag();

    if (here != there   ||   here == room::UNDEFINED) {
      // Here if the two rooms are on different paths (i.e. if they
      // have different flag ID); or, if they have the same ID, if
      // this is UNDEFINED (i.e. they are two unvisited rooms).  In
      // all of these cases, we have to break down the wall.

      pW->setType(wall::OPEN);

      if (here == there) {
        // Two unvisited rooms; create a new path, stored into
        // vPath[nextEntry]

        path rl;
        vPath.push_back(rl);

        vPath[nextEntry].push_back(pBef);
        vPath[nextEntry].push_back(pAft);

        pBef->setFlag(nextEntry);
        pAft->setFlag(nextEntry);
        ++nextEntry;

      } else if (here == room::UNDEFINED) {
        // "Here" is unvisited, "there" visited; add "here" to
        // "there"'s path

        vPath[there].push_back(pBef);
        pBef->setFlag(there);

      } else if (there == room::UNDEFINED) {
        // "Here" is visited, "there" unvisited; add "there" to
        // "here"'s path

        vPath[here].push_back(pAft);
        pAft->setFlag(here);

      } else {
        // Two visited rooms on different paths: merge the path of
        // "here" into the path of "there", and clears the removed
        // path

        path & from = vPath[here];
        path & to   = vPath[there];
        path::iterator it, finis = from.end();

        for (it = from.begin();   it != finis;   ++it) {
          room *pR = *it;
          pR->setFlag(there);
          to.push_back(pR);
        }
        from.clear();
      }
    }
  }
}
