//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Maze generation: class definitions
//
// -------------------------------------------------------------------
//
// $Id: maze.hh,v 1.1 2001/10/05 11:57:40 loreti Exp $

#if !defined(MAZE_HH_)
#define MAZE_HH_

#include <utility>
#include <vector>

class wall;

// This file applies only to rectangular mazes (square rooms).  The
// maze is essentially a graph, and as such may be generated using
// some of the standard algorithms known to find paths in a graph;
// here we define methods for the Depth-First Search (DFS) algorithm,
// and for the Prim and Kruskal algorithms.  All of them are used to
// build a "perfect" maze: i.e. a maze where there is one and only one
// path connecting any pair of rooms.

// A room: abstracted with 2 coordinates, 4 walls and a status (that
// in turn depends from the generation method).  The coordinate system
// is:
//
//    |
//  --+--0--1--2--3---->   (x: width columns of rooms, from 0 to width-1)
//    |
//    0  X  X  X  X  ...
//    |
//    1  X  X  X  X  ...
//    |
//    2  X  X  X  X  ...
//    |
//    |  ...
//    V
//
//  (y: height lines of rooms, from 0 to height-1)

class room {
  friend class maze;

public:
  enum roomStatus { UNDEFINED = -1 };

private:
  unsigned int    _x;
  unsigned int    _y;
  wall         * _upWall;
  wall         * _downWall;
  wall         * _leftWall;
  wall         * _rightWall;
  int            _flag;

  // The status flag may be changed only from the (friend) class maze
  // methods

  void setFlag(int i) { _flag = i; }

public:
  room(unsigned int x = 0, unsigned int y = 0,
       wall * u = 0, wall * d = 0, wall * l = 0, wall * r = 0)
    : _x(x), _y(y), _upWall(u), _downWall(d), _leftWall(l), _rightWall(r),
      _flag(UNDEFINED) {}
  ~room() {}

  int            getFlag()  const { return _flag; }
  unsigned int   getX()     const { return _x; }
  unsigned int   getY()     const { return _y; }
  wall         * getUp()    const { return _upWall; }
  wall         * getDown()  const { return _downWall; }
  wall         * getLeft()  const { return _leftWall; }
  wall         * getRight() const { return _rightWall; }

}; // class room

typedef std::pair<room *, wall *> neighbor;

// A wall: type, and pointers to the two rooms on the sides.

class wall {
  friend class maze;

public:
  enum wallType { ERROR=-1, BORDER, CLOSE, OPEN };

private:
  wallType   _type;
  room     * _before;
  room     * _after;

  // The wall status may be changed only from the (friend) class maze
  // methods

  void setType(wallType t) { _type = t; }

public:
  wall(wallType t = ERROR, room * b = 0, room * a = 0)
    : _type(t), _before(b), _after(a) {}
  ~wall() {}

  wallType   getType()   const { return _type; }
  room     * getBefore() const { return _before; }
  room     * getAfter()  const { return _after; }

}; // class wall

// A maze: the dimensions, a few heavily used constants, and arrays to
// host all the rooms and walls.  The descriptive data are clearly
// redundant, as a (reduced) adjacency matrix is sufficient to define
// the graph.
//
// In an instantiation of the class "maze", the rooms are arranged in
// a dynamic 1-dim array, starting from (0,0) and with the x
// coordinate changing first.  The walls are also arranged in a
// dynamic 1-dim array, and ordered from left to right and from top to
// bottom, as they appear in the following ASCII art (drawn for width
// = 4 and height = 4):
//
//       0   1   2   3
//     +---+---+---+---+    <- "width"   walls
//   0 |   |   |   |   |    <- "width+1" walls
//     +---+---+---+---+    <- "width"   walls
//   1 |   |   |   |   |    <- "width+1" walls
//     +---+---+---+---+
//   2 |   |   |   |   |    That explains the return statements from
//     +---+---+---+---+    the four inspector methods that return the
//   3 |   |   |   |   |    addresses of the four walls of a room,
//     +---+---+---+---+    given its x and y coordinates ("rowFac" is
//                          2*_width+1).

class maze {
public:
  enum buildMethod { DFS, PRIM, KRUSKAL };
  enum entrySide   { HORIZONTAL, VERTICAL };

private:
  int    _width;
  int    _height;
  int    _rowFac;
  int    _roomDim;
  int    _wallDim;
  wall * _wall;
  room * _room;

  // Prevent any use of copy constructor an assignment operator

  maze(const maze &);
  maze & operator = (const maze &);

  // Build methods (DFS, Prim and Kruskal algorithms); and methods
  // internally used by DFS and Prim

  void buildD();
  void buildP();
  void buildK();

  room * DFSlookup(room *, std::vector<room *> *);
  void   primSetNeighbor(room *, std::vector<room *> *);
  void   primGetNeighbor(room *, wall *, std::vector<neighbor> &,
                         std::vector<neighbor> &);

public:
  maze(unsigned int, unsigned int);
  virtual ~maze() { delete [] _wall;  delete [] _room; }

  room * getRoom(int x, int y) const {
    if (x < 0  ||  x >= _width  ||  y < 0  ||  y >= _height) return 0;
    return (_room + (y * _width + x));
  }

  wall * getUpWall(unsigned int x, unsigned int y) const {
    return (_wall + (_rowFac * y + x)); }
  wall * getDownWall(unsigned int x, unsigned int y) const {
    return (_wall + (_rowFac * (y+1) + x)); }
  wall * getLeftWall(unsigned int x, unsigned int y) const {
    return (_wall + (_rowFac * y + _width + x)); }
  wall * getRightWall(unsigned int x, unsigned int y) const {
    return (_wall + (_rowFac * y + _width + x + 1)); }

  void build(const buildMethod, const entrySide);
  void print() const;
};  // class maze

#endif // MAZE_HH_
